#
# Copyright (C) 2007 Red Hat, Inc.
# Authors:
# Thomas Woerner <twoerner@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; version 2 only
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

import os, os.path
import tempfile
import shutil
import types

from fw_config import CHAIN_NAME, MARK_OF_THE_MASQ
from fw_functions import *

##############################################################################

class ip4tablesConfig:
    def __init__(self, filename):
        self.filename = filename
        self.clear()

    def clear(self):
        self.p_config = { }

    def get(self, key):
        if key in self.p_config.keys():
            return self.p_config[key]
        else:
            raise AttributeError, "key not found"

    def set(self, key, value):
        if key not in self.p_config.keys():
            raise AttributeError, "key not found"
        if key[-8:] == "_MODULES":
            self.p_config[key.strip()] = value
        else:
            self.p_config[key.strip()] = value.strip()

    def __str__(self):
        s = ""
        for (key,value) in self.p_config.items():
            if s:
                s += '\n'
            s += '%s = %s' % (key, value)
        return s

    # load self.filename
    def read(self):
        self.clear()
        file = open(self.filename, "r")
        for line in file.xreadlines():
            if not line:
                break
            line = line.strip()
            if len(line) < 1 or line[0] == '#':
                continue
            # get key/value pairs
            p = line.split("=")
            if len(p) != 2:
                continue
            key = p[0].strip()
            value = p[1].strip()
            # remove leading and trailing double quotes
            if value[0] == '"' and value[-1] == '"':
                value = value[1:-1]
            if key[-8:] == "_MODULES":
                value = value.split()
            self.p_config[key] = value
        file.close()

    # save to self.filename if there are key/value changes
    def write(self):
        if len(self.p_config) < 1:
            # no changes: nothing to do
            return

        (temp_file, temp) = tempfile.mkstemp("iptables-config")
        modified = False
        file = open(self.filename, "r")
        for line in file.xreadlines():
            if not line: break
            # remove newline at and on line
            if line[-1:] == "\n":
                line = line[:-1]
            if len(line) < 1:
                os.write(temp_file, "\n")
                continue

            if line[0] != "#" and len(line) > 1:
                p = line.split("=")
                if len(p) != 2:
                    os.write(temp_file, line+"\n")
                    continue
                key = p[0].strip()
                value = p[1].strip()
                # remove leading and trailing double quotes
                if value[0] == '"' and value[-1] == '"':
                    value = value[1:-1]
                if key[-8:] == "_MODULES":
                    value = value.split()
                if key in self.p_config.keys() and self.p_config[key] != value:
                    self._write(temp_file, key, self.p_config[key])
                    modified = True
                    del self.p_config[key]
                else:
                    os.write(temp_file, line+"\n")
                    del self.p_config[key]
            else:
                os.write(temp_file, line+"\n")
        # write remaining key/value pairs
        if len(self.p_config) > 0:
            os.write(temp_file, "\n")
        for (key,value) in self.p_config.items():
            self._write(temp_file, key, value)
            modified = True

        file.close()
        os.close(temp_file)

        if not modified: # not modified: remove tempfile
            os.remove(temp)
            return
        # copy tempfile
        try:
            shutil.copy(temp, self.filename)
        except:
            os.remove(temp)
            raise IOError, "Permission denied: '%s'" % self.filename
        else:
            os.remove(temp)
            os.chmod(self.filename, 0600)

    def _write(self, fd, key, value):
        if isinstance(value, types.ListType) or \
               isinstance(value, types.TupleType):
            val = " ".join(value)
        else:
            val = value
        os.write(fd, '%s="%s"\n' % (key, val))

##############################################################################

ip6tablesConfig = ip4tablesConfig

##############################################################################

class iptablesClass:
    def __init__(self, filename):
        self.filename = filename

    def write(self, conf):
        custom_mangle = [ ]
        custom_nat = [ ]
        custom_filter = [ ]
        if conf.custom_rules and len(conf.custom_rules) > 0:
            for (type, table, _filename) in conf.custom_rules:
                if type != "ipv4":
                    continue
                if table == "mangle":
                    custom_mangle.append(_filename)
                elif table == "nat":
                    custom_nat.append(_filename)
                elif table == "filter":
                    custom_filter.append(_filename)

        fd = open(self.filename, "w")

        fd.write("# Firewall configuration written by system-config-firewall\n")
        fd.write("# Manual customization of this file is not recommended.\n")

        # mangle table
        if (conf.masq and len(conf.masq) > 0) or len(custom_mangle) > 0:
            fd.write("*mangle\n")
            fd.write(":PREROUTING ACCEPT [0:0]\n")
            fd.write(":INPUT ACCEPT [0:0]\n")
            fd.write(":FORWARD ACCEPT [0:0]\n")
            fd.write(":OUTPUT ACCEPT [0:0]\n")
            fd.write(":POSTROUTING ACCEPT [0:0]\n")
            # masuerading devices
            if (conf.masq and len(conf.masq) > 0):
                for dev in conf.masq:
                    fd.write("-A PREROUTING -i %s -j MARK --set-mark 0x%X\n" % \
                             (dev, MARK_OF_THE_MASQ))
            # add custom mangle rules
            if len(custom_mangle) > 0:
                for _filename in custom_mangle:
                    catFile(fd, _filename)
            fd.write("COMMIT\n")

        # nat table
        if (conf.masq and len(conf.masq) > 0) or len(custom_nat) > 0:
            fd.write("*nat\n")
            fd.write(":PREROUTING ACCEPT [0:0]\n")
            fd.write(":OUTPUT ACCEPT [0:0]\n")
            fd.write(":POSTROUTING ACCEPT [0:0]\n")
            # masquerade
            if (conf.masq and len(conf.masq) > 0):
                fd.write("-A POSTROUTING -m mark --mark 0x%X "
                         "-j MASQUERADE\n" % MARK_OF_THE_MASQ)
            # add custom nat rules
            if len(custom_nat) > 0:
                for _filename in custom_nat:
                    catFile(fd, _filename)
            fd.write("COMMIT\n")

        fd.write("*filter\n")
        fd.write(":INPUT ACCEPT [0:0]\n")
        fd.write(":FORWARD ACCEPT [0:0]\n")
        fd.write(":OUTPUT ACCEPT [0:0]\n")
        fd.write(":%s - [0:0]\n" % CHAIN_NAME)
        fd.write("-A INPUT -j %s\n" % CHAIN_NAME)

        # accept lo
        fd.write("-A %s -i lo -j ACCEPT\n" % CHAIN_NAME)

        if conf.trust:
            for dev in conf.trust:
                fd.write("-A %s -i %s -j ACCEPT\n" % (CHAIN_NAME, dev))

        # allow ICMP
        fd.write("-A %s -p icmp --icmp-type any -j ACCEPT\n" % CHAIN_NAME)

        # allow IPSec traffic (matches Lokkit 0.50 behavior)
        if not conf.no_ipsec:
            fd.write("-A %s -p 50 -j ACCEPT\n" % CHAIN_NAME)
            fd.write("-A %s -p 51 -j ACCEPT\n" % CHAIN_NAME)

        # allow mDNS
        if not conf.no_mdns:
            fd.write("-A %s -p udp --dport 5353 -d 224.0.0.251 -j ACCEPT\n" % \
                     CHAIN_NAME)

        # allow CUPS browsing
        if not conf.no_ipp:
            fd.write("-A %s -p udp -m udp --dport 631 -j ACCEPT\n" % CHAIN_NAME)
            fd.write("-A %s -p tcp -m tcp --dport 631 -j ACCEPT\n" % CHAIN_NAME)

        # note: RELATED is extremely important as it matches ICMP error messages
        fd.write("-A %s -m state --state ESTABLISHED,RELATED -j ACCEPT\n" % \
                 CHAIN_NAME)

        # open ports for new connections
        if conf.ports and len(conf.ports) > 0:
            for (ports, proto) in conf.ports:
                if len(ports) == 1:
                    dports = "%s" % ports[0]
                else:
                    dports = "%s:%s" % (ports[0], ports[1])
                fd.write("-A %s -m state --state NEW -m %s -p %s --dport %s "
                         "-j ACCEPT\n" % (CHAIN_NAME, proto, proto, dports))

        # add custom filter rules
        if len(custom_filter) > 0:
            for _filename in custom_filter:
                catFile(fd, _filename)

        # reject all other
        reject_type = "icmp-host-prohibited" # icmp-admin-prohibited
        fd.write("-A %s -j REJECT --reject-with %s\n" % (CHAIN_NAME,
                                                         reject_type))
        fd.write("-A FORWARD -j REJECT --reject-with %s\n" % reject_type)

        fd.write("COMMIT\n");
        fd.close()

    def start(self):
        return os.system("/sbin/service iptables start >/dev/null 2>&1")

    def restart(self):
        return os.system("/sbin/service iptables restart >/dev/null 2>&1")

    def stop(self):
        return os.system("/sbin/service iptables stop >/dev/null 2>&1")

    def chkconfig_on(self):
        return os.system("/sbin/chkconfig iptables on >/dev/null 2>&1")

    def chkconfig_off(self):
        return os.system("/sbin/chkconfig iptables off >/dev/null 2>&1")

    def status(self):
        return os.system("/sbin/chkconfig iptables status >/dev/null 2>&1")

    def unlink(self):
        if os.path.exists(self.filename) and os.path.isfile(self.filename):
            os.unlink(self.filename)

##############################################################################

class ip6tablesClass:
    def __init__(self, filename):
        self.filename = filename

    def write(self, conf):
        custom_mangle = [ ]
        custom_nat = [ ]
        custom_filter = [ ]
        if conf.custom_rules and len(conf.custom_rules) > 0:
            for (type, table, _filename) in conf.custom_rules:
                if type != "ipv6":
                    continue
                if table == "mangle":
                    custom_mangle.append(_filename)
                elif table == "nat":
                    custom_nat.append(_filename)
                elif table == "filter":
                    custom_filter.append(_filename)

        fd = open(self.filename, "w")

        fd.write("# Firewall configuration written by system-config-firewall\n")
        fd.write("# Manual customization of this file is not recommended.\n")

        # mangle table
# no support for masquerading for netfilterv6 for now
#        if (conf.masq and len(conf.masq) > 0) or len(custom_mangle) > 0:
        if len(custom_mangle) > 0:
            fd.write("*mangle\n")
            fd.write(":PREROUTING ACCEPT [0:0]\n")
            fd.write(":INPUT ACCEPT [0:0]\n")
            fd.write(":FORWARD ACCEPT [0:0]\n")
            fd.write(":OUTPUT ACCEPT [0:0]\n")
            fd.write(":POSTROUTING ACCEPT [0:0]\n")
# no support for masquerading for netfilterv6 for now
#            # masuerading devices
#            if (conf.masq and len(conf.masq) > 0):
#                for dev in conf.masq:
#                    fd.write("-A PREROUTING -i %s -j MARK --set-mark 0x%X\n" % \
#                             (dev, MARK_OF_THE_MASQ))
            # add custom mangle rules
            if len(custom_mangle) > 0:
                for _filename in custom_mangle:
                    catFile(fd, _filename)
            fd.write("COMMIT\n")

# no support for nat for netfilterv6 for now
#        # nat table
#        if (conf.masq and len(conf.masq) > 0) or len(custom_nat) > 0:
#            fd.write("*nat\n")
#            fd.write(":PREROUTING ACCEPT [0:0]\n")
#            fd.write(":OUTPUT ACCEPT [0:0]\n")
#            fd.write(":POSTROUTING ACCEPT [0:0]\n")
#            # masquerade
#            if (conf.masq and len(conf.masq) > 0):
#                fd.write("-A POSTROUTING -m mark --mark 0x%X "
#                         "-j MASQUERADE\n" % MARK_OF_THE_MASQ)
#            # add custom nat rules
#            if len(custom_nat) > 0:
#                for _filename in custom_nat:
#                    catFile(fd, _filename)
#            fd.write("COMMIT\n")

        fd.write("*filter\n")
        fd.write(":INPUT ACCEPT [0:0]\n")
        fd.write(":FORWARD ACCEPT [0:0]\n")
        fd.write(":OUTPUT ACCEPT [0:0]\n")
        fd.write(":%s - [0:0]\n" % CHAIN_NAME)
        fd.write("-A INPUT -j %s\n" % CHAIN_NAME)

        # accept lo
        fd.write("-A %s -i lo -j ACCEPT\n" % CHAIN_NAME)

        if conf.trust:
            for dev in conf.trust:
                fd.write("-A %s -i %s -j ACCEPT\n" % (CHAIN_NAME, dev))

        # allow ICMP
        fd.write("-A %s -p icmpv6 -j ACCEPT\n" % CHAIN_NAME)

        # allow IPSec traffic (matches Lokkit 0.50 behavior)
        if not conf.no_ipsec:
            fd.write("-A %s -p 50 -j ACCEPT\n" % CHAIN_NAME)
            fd.write("-A %s -p 51 -j ACCEPT\n" % CHAIN_NAME)

        # allow mDNS
        if not conf.no_mdns:
            fd.write("-A %s -p udp --dport 5353 -d ff02::fb -j ACCEPT\n" % \
                     CHAIN_NAME)

        # allow CUPS browsing
        if not conf.no_ipp:
            fd.write("-A %s -p udp -m udp --dport 631 -j ACCEPT\n" % CHAIN_NAME)
            fd.write("-A %s -p tcp -m tcp --dport 631 -j ACCEPT\n" % CHAIN_NAME)

        # note: RELATED is extremely important as it matches ICMP error messages
        fd.write("-A %s -m state --state ESTABLISHED,RELATED -j ACCEPT\n" % \
                 CHAIN_NAME)

# use state module - this method is not secure
#        # Allow outgoing connections
#        fd.write("-A %s -p udp -m udp --dport 32768:61000 -j ACCEPT\n" % \
#                 CHAIN_NAME)
#        fd.write("-A %s -p tcp -m tcp --dport 32768:61000 -j ACCEPT\n" % \
#                 CHAIN_NAME)

        # open ports for new connections
        if conf.ports and len(conf.ports) > 0:
            for (ports, proto) in conf.ports:
                if len(ports) == 1:
                    dports = "%s" % ports[0]
                else:
                    dports = "%s:%s" % (ports[0], ports[1])
                fd.write("-A %s -m state --state NEW -m %s -p %s --dport %s "
                         "-j ACCEPT\n" % (CHAIN_NAME, proto, proto, dports))

        # add custom filter rules
        if len(custom_filter) > 0:
            for _filename in custom_filter:
                catFile(fd, _filename)

        # reject all other
        reject_type = "icmp6-port-unreachable" # icmp6-adm-prohibited
        fd.write("-A %s -j REJECT --reject-with %s\n" % (CHAIN_NAME,
                                                         reject_type))
        fd.write("-A FORWARD -j REJECT --reject-with %s\n" % reject_type)

        fd.write("COMMIT\n");
        fd.close()

    def start(self):
        return os.system("/sbin/service ip6tables start >/dev/null 2>&1")

    def restart(self):
        return os.system("/sbin/service ip6tables restart >/dev/null 2>&1")

    def stop(self):
        return os.system("/sbin/service ip6tables stop >/dev/null 2>&1")

    def chkconfig_on(self):
        return os.system("/sbin/chkconfig ip6tables on >/dev/null 2>&1")

    def chkconfig_off(self):
        return os.system("/sbin/chkconfig ip6tables off >/dev/null 2>&1")

    def status(self):
        return os.system("/sbin/chkconfig ip6tables status >/dev/null 2>&1")

    def unlink(self):
        if os.path.exists(self.filename) and os.path.isfile(self.filename):
            os.unlink(self.filename)
