#
# Copyright (C) 2007 Red Hat, Inc.
# Authors:
# Thomas Woerner <twoerner@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; version 2 only
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

from copy import copy
from optparse import Option, OptionValueError, OptionParser, Values
import socket
from fw_config import FIREWALL_TYPES, FIREWALL_TABLES
from fw_functions import getPortID, getPortRange, getServiceName
import os.path

def _check_selinuxmode(option, opt, value):
    if value not in [ "enforcing", "permissive", "disabled" ]:
        raise OptionValueError("option %s: invalid selinux mode '%s'." % \
                               (opt, value))
    return value

def _check_port(option, opt, value):
    failure = False
    try:
        (ports, protocol) = value.split(":")
    except:
        failure = True
    if not failure:
        range = getPortRange(ports.strip())
        if range == -1:
            failure = True
        elif range == None:
            raise OptionValueError("option %s: port range '%s' is not "
                                   "unique." % (opt, value))
        elif len(range) == 2 and range[0] >= range[1]:
            raise OptionValueError("option %s: '%s' is not a valid range "
                                   "(start port >= end port)." % (opt, value))
    if not failure:
        protocol = protocol.strip()
        if protocol not in [ "tcp", "udp" ]:
            raise OptionValueError("option %s: '%s' is not a valid protocol." \
                                   % (opt, protocol))
    if failure:
        raise OptionValueError("option %s: invalid port definition '%s'." % \
                               (opt, value))
    return (range, protocol)

def _check_rulesfile(option, opt, value):
    type = "ipv4"
    table = "filter"
    splits = value.split(":", 1)
    if len(splits) > 1 and splits[0] in FIREWALL_TYPES:
        type = splits[0]
        splits = splits[1].split(":", 1)
    if len(splits) > 1 and splits[0] in FIREWALL_TABLES:
        table = splits[0]
        splits = splits[1].split(":", 1)
    filename = ":".join(splits)

    if type == "ipv6" and table == "nat":
        raise OptionValueError("option %s=%s: ipv6 has no nat support "
                               "currently." % (opt, value))
    
    if os.path.exists(filename) and os.path.isfile(filename):
        return (type, table, filename)

    raise OptionValueError("option %s=%s: file '%s' does not exist." % \
                           (opt, value, filename))

class _Option(Option):
    TYPES = Option.TYPES + ("port", "rulesfile", "selinuxmode", )
    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
    TYPE_CHECKER["port"] = _check_port
    TYPE_CHECKER["rulesfile"] = _check_rulesfile
    TYPE_CHECKER["selinuxmode"] = _check_selinuxmode

def _addStandardOptions(parser):
    parser.add_option("--enabled",
                      action="store_true", dest="enabled",
                      help=_("Enable firewall (default)"))
    parser.add_option("--disabled",
                      action="store_false", dest="enabled",
                      help=_("Disable firewall"))
    parser.add_option("--addmodule",
                      action="append", dest="add_module", type="string",
                      metavar=_("<module>"),
                      help=_("Enable an iptables module"))
    parser.add_option("--removemodule",
                      action="append", dest="remove_module", type="string",
                      metavar=_("<module>"),
                      help=_("Disable an iptables module"))
    parser.add_option("-p", "--port",
                      action="append", dest="ports",
                      metavar=_("<port>[-<port>]:<protocol>"),
                      type="port",
                      help=_("Allow specific ports through the firewall "
                             "(e.g, ssh:tcp)"))
    parser.add_option("--no-ipsec",
                      action="store_true", dest="no_ipsec",
                      help=_("Disable IPSec"))
    parser.add_option("--no-ipp",
                      action="store_true", dest="no_ipp",
                      help=_("Disable Internet Printing Protocol"))
    parser.add_option("--no-mdns",
                      action="store_true", dest="no_mdns",
                      help=_("Disable Multicast DNS"))
    parser.add_option("-t", "--trust",
                      action="append", dest="trust", type="string",
                      metavar=_("<interface>"),
                      help=_("Allow all traffic on the specified device"))
    parser.add_option("-m", "--masq",
                      action="append", dest="masq",
                      metavar=_("<interface>"),
                      help=_("Masquerades traffic from the specified device"))
    parser.add_option( "--high", "--medium",
                      action="store_true", dest="enabled",
                      help=_("Backwards compatibilty, aliased to --enabled"))
    parser.add_option("--custom-rules",
                      action="append", dest="custom_rules", type="rulesfile",
                      metavar=_("[<type>:][<table>:]<filename>"),
                      help=_("Specify a custom rules file for inclusion in "
                             "the firewall. It will be included after the "
                             "default rules. (e.g, ipv4:filter:/etc/sysconfig/"
                             "ipv4_filter_addon)"))

def _addSELinuxOptions(parser):
    parser.add_option("--selinux",
                      action="store", dest="selinux", type="selinuxmode",
                      metavar=_("<mode>"),
                      help=_("Configure SELinux mode: enforcing, permissive "
                             "or disabled"))
    parser.add_option("--selinuxtype",
                      action="store", dest="selinuxtype", type="string",
                      metavar=_("<type>"),
                      help=_("Configure SELinux type: Usually targeted or "
                             "strict Policy"))

def _parse_args(parser, args, options=None):
    (_options, _args) = parser.parse_args(args, options)
    if len(_args) != 0:
        parser.error(_("no such option: %s") % " ".join(_args))
    return _options

def _gen_parser():
    parser = OptionParser(add_help_option=False, option_class=_Option)
    # always print full usage message
    parser.print_usage = parser.print_help
    return parser

def parseSysconfigArgs(args, options=None):
    parser = _gen_parser()
    _addStandardOptions(parser)
    return _parse_args(parser, args, options)

def parseSELinuxArgs(args, options=None):
    parser = _gen_parser()
    _addSELinuxOptions(parser)
    return _parse_args(parser, args, options)

def parseArgs(args=None, options=None):
    parser = _gen_parser()
    parser.add_option("-?", "--help", "--usage", action="help",
                      help=_("Show this help message"))
    parser.add_option("-q", "--quiet",
                      action="store_true", dest="quiet",
                      help=_("Run noninteractively; process only command-line "
                             "arguments"))
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose",
                      help=_("Be more verbose"))
    parser.add_option("-n", "--nostart",
                      action="store_true", dest="nostart",
                      help=_("Configure firewall but do not activate it"))
    parser.add_option("-f",
                      action="store_true", dest="force",
                      help=_("Ignore actual settings"))
    parser.add_option("--stop",
                      action="store_true", dest="stop",
                      help=_("Stop firewall if firewall gets disabled"))
    _addSELinuxOptions(parser)
    _addStandardOptions(parser)

    return _parse_args(parser, args, options)

def copyValues(values):
    new_values = Values()
    new_values.__dict__ = copy(values.__dict__)
    return new_values
